@@ -29,10 +29,10 @@ from __future__ import absolute_import, unicode_literals
 
 from . import cbor
 from .ctap import CtapError
-from .cose import CoseKey
+from .cose import CoseKey, ES256
 from .hid import CTAPHID, CAPABILITY
 from .utils import Timeout, sha256, hmac_sha256, bytes2int, int2bytes
-from .attestation import Attestation, FidoU2FAttestation
+from .attestation import FidoU2FAttestation
 
 from cryptography.hazmat.backends import default_backend
 from cryptography.hazmat.primitives.asymmetric import ec
@@ -64,13 +64,6 @@ def hexstr(bs):
     return "h'%s'" % b2a_hex(bs).decode()
 
 
-def _parse_cbor(data):
-    resp, rest = cbor.loads(data)
-    if rest:
-        raise ValueError('Extraneous data')
-    return resp
-
-
 class Info(bytes):
     """Binary CBOR encoded response data returned by the CTAP2 GET_INFO command.
 
@@ -87,17 +80,29 @@ class Info(bytes):
 
     @unique
     class KEY(IntEnum):
-        VERSIONS = 1
-        EXTENSIONS = 2
-        AAGUID = 3
-        OPTIONS = 4
-        MAX_MSG_SIZE = 5
-        PIN_PROTOCOLS = 6
+        VERSIONS = 0x01
+        EXTENSIONS = 0x02
+        AAGUID = 0x03
+        OPTIONS = 0x04
+        MAX_MSG_SIZE = 0x05
+        PIN_PROTOCOLS = 0x06
+        MAX_CREDS_IN_LIST = 0x07
+        MAX_CRED_ID_LENGTH = 0x08
+        TRANSPORTS = 0x09
+        ALGORITHMS = 0x0a
+
+        @classmethod
+        def get(cls, key):
+            try:
+                return cls(key)
+            except ValueError:
+                return key
 
     def __init__(self, _):
         super(Info, self).__init__()
 
-        data = dict((Info.KEY(k), v) for (k, v) in _parse_cbor(self).items())
+        data = dict((Info.KEY.get(k), v) for (k, v) in
+                    cbor.decode(self).items())
         self.versions = data[Info.KEY.VERSIONS]
         self.extensions = data.get(Info.KEY.EXTENSIONS, [])
         self.aaguid = data[Info.KEY.AAGUID]
@@ -105,6 +110,10 @@ class Info(bytes):
         self.max_msg_size = data.get(Info.KEY.MAX_MSG_SIZE, 1024)
         self.pin_protocols = data.get(
             Info.KEY.PIN_PROTOCOLS, [])
+        self.max_creds_in_list = data.get(Info.KEY.MAX_CREDS_IN_LIST)
+        self.max_cred_id_length = data.get(Info.KEY.MAX_CRED_ID_LENGTH)
+        self.transports = data.get(Info.KEY.TRANSPORTS, [])
+        self.algorithms = data.get(Info.KEY.ALGORITHMS)
         self.data = data
 
     def __repr__(self):
@@ -117,6 +126,14 @@ class Info(bytes):
         r += ', max_message_size: %d' % self.max_msg_size
         if self.pin_protocols:
             r += ', pin_protocols: %r' % self.pin_protocols
+        if self.max_creds_in_list:
+            r += ', max_credential_count_in_list: %d' % self.max_creds_in_list
+        if self.max_cred_id_length:
+            r += ', max_credential_id_length: %d' % self.max_cred_id_length
+        if self.transports:
+            r += ', transports: %r' % self.transports
+        if self.algorithms:
+            r += ', algorithms: %r' % self.algorithms
         return r + ')'
 
     def __str__(self):
@@ -162,7 +179,7 @@ class AttestedCredentialData(bytes):
         aaguid = data[:16]
         c_len = struct.unpack('>H', data[16:18])[0]
         cred_id = data[18:18+c_len]
-        pub_key, rest = cbor.loads(data[18+c_len:])
+        pub_key, rest = cbor.decode_from(data[18+c_len:])
         return aaguid, cred_id, CoseKey.parse(pub_key), rest
 
     @classmethod
@@ -175,7 +192,7 @@ class AttestedCredentialData(bytes):
         :return: The attested credential data.
         """
         return cls(aaguid + struct.pack('>H', len(credential_id))
-                   + credential_id + cbor.dumps(public_key))
+                   + credential_id + cbor.encode(public_key))
 
     @classmethod
     def unpack_from(cls, data):
@@ -189,6 +206,24 @@ class AttestedCredentialData(bytes):
         parts = cls.parse(data)
         return cls.create(*parts[:-1]), parts[-1]
 
+    @classmethod
+    def from_ctap1(cls, key_handle, public_key):
+        """Create an AttestatedCredentialData from a CTAP1 RegistrationData
+        instance.
+
+        :param key_handle: The CTAP1 credential key_handle.
+        :type key_handle: bytes
+        :param public_key: The CTAP1 65 byte public key.
+        :type public_key: bytes
+        :return: The credential data, using an all-zero AAGUID.
+        :rtype: AttestedCredentialData
+        """
+        return cls.create(
+            b'\0'*16,  # AAGUID
+            key_handle,
+            ES256.from_ctap1(public_key)
+        )
+
 
 class AuthenticatorData(bytes):
     """Binary encoding of the authenticator data.
@@ -204,8 +239,7 @@ class AuthenticatorData(bytes):
 
     @unique
     class FLAG(IntEnum):
-        """
-        Authenticator data flags
+        """Authenticator data flags
 
         See https://www.w3.org/TR/webauthn/#sec-authenticator-data for details
         """
@@ -230,7 +264,7 @@ class AuthenticatorData(bytes):
             self.credential_data = None
 
         if self.flags & AuthenticatorData.FLAG.EXTENSION_DATA:
-            self.extensions, rest = cbor.loads(rest)
+            self.extensions, rest = cbor.decode_from(rest)
         else:
             self.extensions = None
 
@@ -252,7 +286,7 @@ class AuthenticatorData(bytes):
         """
         return cls(
             rp_id_hash + struct.pack('>BI', flags, counter) + credential_data +
-            (cbor.dumps(extensions) if extensions is not None else b'')
+            (cbor.encode(extensions) if extensions is not None else b'')
         )
 
     def is_user_present(self):
@@ -350,7 +384,7 @@ class AttestationObject(bytes):
         super(AttestationObject, self).__init__()
 
         data = dict((AttestationObject.KEY.for_key(k), v) for (k, v) in
-                    _parse_cbor(self).items())
+                    cbor.decode(self).items())
         self.fmt = data[AttestationObject.KEY.FMT]
         self.auth_data = AuthenticatorData(data[AttestationObject.KEY.AUTH_DATA]
                                            )
@@ -366,20 +400,6 @@ class AttestationObject(bytes):
     def __str__(self):
         return self.__repr__()
 
-    def verify(self, client_param):
-        """Verify the digital signature of an AttestationObject, with regard to
-        the given client_param.
-
-        :param client_param: SHA256 hash of the ClientData for the request.
-        :type client_param: bytes
-        """
-        attestation = Attestation.for_type(self.fmt)
-        if attestation:
-            attestation().verify(self.att_statement, self.auth_data,
-                                 client_param)
-        else:
-            raise ValueError('Unsupported format: %s' % self.fmt)
-
     @classmethod
     def create(cls, fmt, auth_data, att_stmt):
         """Create an AttestationObject instance.
@@ -393,7 +413,7 @@ class AttestationObject(bytes):
         :return: The attestation object.
         :rtype: AttestationObject
         """
-        return cls(cbor.dumps(args(fmt, auth_data, att_stmt)))
+        return cls(cbor.encode(args(fmt, auth_data, att_stmt)))
 
     @classmethod
     def from_ctap1(cls, app_param, registration):
@@ -412,16 +432,9 @@ class AttestationObject(bytes):
                 app_param,
                 0x41,
                 0,
-                AttestedCredentialData.create(
-                    b'\0'*16,  # aaguid
+                AttestedCredentialData.from_ctap1(
                     registration.key_handle,
-                    {  # EC256 public key
-                        1: 2,
-                        3: -7,
-                        -1: 1,
-                        -2: registration.public_key[1:1+32],
-                        -3: registration.public_key[33:33+32]
-                    }
+                    registration.public_key
                 )
             ),
             {  # att_statement
@@ -437,7 +450,7 @@ class AttestationObject(bytes):
         :return: The attestation object, using int keys.
         :rtype: AttestationObject
         """
-        return AttestationObject(cbor.dumps(self.data))
+        return AttestationObject(cbor.encode(self.data))
 
     def with_string_keys(self):
         """Get a copy of this AttestationObject, using Webauthn string values as
@@ -446,7 +459,7 @@ class AttestationObject(bytes):
         :return: The attestation object, using str keys.
         :rtype: AttestationObject
         """
-        return AttestationObject(cbor.dumps(
+        return AttestationObject(cbor.encode(
             dict((k.string_key, v) for k, v in self.data.items())))
 
 
@@ -474,7 +487,7 @@ class AssertionResponse(bytes):
         super(AssertionResponse, self).__init__()
 
         data = dict((AssertionResponse.KEY(k), v) for (k, v) in
-                    _parse_cbor(self).items())
+                    cbor.decode(self).items())
         self.credential = data.get(
             AssertionResponse.KEY.CREDENTIAL)
         self.auth_data = AuthenticatorData(
@@ -518,8 +531,8 @@ class AssertionResponse(bytes):
         :param n_creds: The number of responses available.
         :return: The assertion response.
         """
-        return cls(cbor.dumps(args(credential, auth_data, signature, user,
-                                   n_creds)))
+        return cls(cbor.encode(args(credential, auth_data, signature, user,
+                                    n_creds)))
 
     @classmethod
     def from_ctap1(cls, app_param, credential, authentication):
@@ -546,6 +559,8 @@ class CTAP2(object):
     """Implementation of the CTAP2 specification.
 
     :param device: A CtapHidDevice handle supporting CTAP2.
+    :param strict_cbor: Validate that CBOR returned from the Authenticator is
+        canonical, defaults to True.
     """
 
     @unique
@@ -556,15 +571,17 @@ class CTAP2(object):
         CLIENT_PIN = 0x06
         RESET = 0x07
         GET_NEXT_ASSERTION = 0x08
+        # 0x41 is the command byte for credmgmt preview
+        CREDENTIAL_MGMT = 0x41
 
-    def __init__(self, device):
+    def __init__(self, device, strict_cbor=True):
         if not device.capabilities & CAPABILITY.CBOR:
             raise ValueError('Device does not support CTAP2.')
         self.device = device
+        self._strict_cbor = strict_cbor
 
-    def send_cbor(self, cmd, data=None, timeout=None, parse=_parse_cbor,
+    def send_cbor(self, cmd, data=None, timeout=None, parse=cbor.decode,
                   on_keepalive=None):
-
         """Sends a CBOR message to the device, and waits for a response.
 
         The optional parameter 'timeout' can either be a numeric time in seconds
@@ -583,7 +600,7 @@ class CTAP2(object):
         """
         request = struct.pack('>B', cmd)
         if data is not None:
-            request += cbor.dumps(data)
+            request += cbor.encode(data)
         with Timeout(timeout) as event:
             response = self.device.call(CTAPHID.CBOR, request, event,
                                         on_keepalive)
@@ -592,13 +609,22 @@ class CTAP2(object):
             raise CtapError(status)
         if len(response) == 1:
             return None
-        return parse(response[1:])
+        enc = response[1:]
+        if self._strict_cbor:
+            expected = cbor.encode(cbor.decode(enc))
+            if expected != enc:
+                enc_h = b2a_hex(enc)
+                exp_h = b2a_hex(expected)
+                raise ValueError('Non-canonical CBOR from Authenticator.\n'
+                                 'Got: {}\n'.format(enc_h) +
+                                 'Expected: {}'.format(exp_h))
+        return parse(enc)
 
     def make_credential(self, client_data_hash, rp, user, key_params,
                         exclude_list=None, extensions=None, options=None,
                         pin_auth=None, pin_protocol=None, timeout=None,
                         on_keepalive=None):
-        """CTAP2 makeCredential operation,
+        """CTAP2 makeCredential operation.
 
         :param client_data_hash: SHA256 hash of the ClientData.
         :param rp: PublicKeyCredentialRpEntity parameters.
@@ -632,7 +658,7 @@ class CTAP2(object):
                       pin_protocol=None, timeout=None, on_keepalive=None):
         """CTAP2 getAssertion command.
 
-        :param rp_id: SHA256 hash of the RP ID of the credential.
+        :param rp_id: The RP ID of the credential.
         :param client_data_hash: SHA256 hash of the ClientData used.
         :param allow_list: Optional list of PublicKeyCredentialDescriptors.
         :param extensions: Optional dict of extensions.
@@ -702,6 +728,33 @@ class CTAP2(object):
         return self.send_cbor(CTAP2.CMD.GET_NEXT_ASSERTION,
                               parse=AssertionResponse)
 
+    def credential_mgmt(self, sub_cmd, sub_cmd_params=None, pin_protocol=None,
+                        pin_auth=None):
+        """CTAP2 credentialManagement command, used to manage resident
+        credentials.
+
+        :param sub_cmd: A credentialManagement sub command.
+        :param sub_cmd_params: Sub command specific parameters.
+        :param pin_protocol: PIN protocol version used.
+        :pin_auth:
+        """
+        return self.send_cbor(CTAP2.CMD.CREDENTIAL_MGMT, args(
+            sub_cmd,
+            sub_cmd_params,
+            pin_protocol,
+            pin_auth
+        ))
+
+    def get_assertions(self, *args, **kwargs):
+        """Convenience method to get list of assertions.
+
+        See get_assertion and get_assertion_next for details.
+        """
+        first = self.get_assertion(*args, **kwargs)
+        rest = [self.get_assertion_next()
+                for _ in range(1, first.number_of_credentials or 1)]
+        return [first] + rest
+
 
 def _pad_pin(pin):
     if not isinstance(pin, six.string_types):
@@ -716,7 +769,7 @@ def _pad_pin(pin):
 
 
 class PinProtocolV1(object):
-    """Implementation of the CTAP1 PIN protocol v1.
+    """Implementation of the CTAP2 PIN protocol v1.
 
     :param ctap: An instance of a CTAP2 object.
     :cvar VERSION: The version number of the PIV protocol.
@@ -742,7 +795,7 @@ class PinProtocolV1(object):
     def __init__(self, ctap):
         self.ctap = ctap
 
-    def _init_shared_secret(self):
+    def get_shared_secret(self):
         be = default_backend()
         sk = ec.generate_private_key(ec.SECP256R1(), be)
         pn = sk.public_key().public_numbers()
@@ -762,17 +815,19 @@ class PinProtocolV1(object):
         shared_secret = sha256(sk.exchange(ec.ECDH(), pk))  # x-coordinate, 32b
         return key_agreement, shared_secret
 
+    def _get_cipher(self, secret):
+        be = default_backend()
+        return Cipher(algorithms.AES(secret), modes.CBC(PinProtocolV1.IV), be)
+
     def get_pin_token(self, pin):
         """Get a PIN token from the authenticator.
 
         :param pin: The PIN of the authenticator.
         :return: A PIN token.
         """
-        key_agreement, shared_secret = self._init_shared_secret()
+        key_agreement, shared_secret = self.get_shared_secret()
 
-        be = default_backend()
-        cipher = Cipher(algorithms.AES(shared_secret),
-                        modes.CBC(PinProtocolV1.IV), be)
+        cipher = self._get_cipher(shared_secret)
         pin_hash = sha256(pin.encode())[:16]
         enc = cipher.encryptor()
         pin_hash_enc = enc.update(pin_hash) + enc.finalize()
@@ -795,17 +850,16 @@ class PinProtocolV1(object):
 
     def set_pin(self, pin):
         """Set the PIN of the autenticator.
+
         This only works when no PIN is set. To change the PIN when set, use
         change_pin.
 
         :param pin: A PIN to set.
         """
         pin = _pad_pin(pin)
-        key_agreement, shared_secret = self._init_shared_secret()
+        key_agreement, shared_secret = self.get_shared_secret()
 
-        be = default_backend()
-        cipher = Cipher(algorithms.AES(shared_secret),
-                        modes.CBC(PinProtocolV1.IV), be)
+        cipher = self._get_cipher(shared_secret)
         enc = cipher.encryptor()
         pin_enc = enc.update(pin) + enc.finalize()
         pin_auth = hmac_sha256(shared_secret, pin_enc)[:16]
@@ -816,6 +870,7 @@ class PinProtocolV1(object):
 
     def change_pin(self, old_pin, new_pin):
         """Change the PIN of the authenticator.
+
         This only works when a PIN is already set. If no PIN is set, use
         set_pin.
 
@@ -823,11 +878,9 @@ class PinProtocolV1(object):
         :param new_pin: The new PIN to set.
         """
         new_pin = _pad_pin(new_pin)
-        key_agreement, shared_secret = self._init_shared_secret()
+        key_agreement, shared_secret = self.get_shared_secret()
 
-        be = default_backend()
-        cipher = Cipher(algorithms.AES(shared_secret),
-                        modes.CBC(PinProtocolV1.IV), be)
+        cipher = self._get_cipher(shared_secret)
         pin_hash = sha256(old_pin.encode())[:16]
         enc = cipher.encryptor()
         pin_hash_enc = enc.update(pin_hash) + enc.finalize()
@@ -840,3 +893,165 @@ class PinProtocolV1(object):
                              pin_hash_enc=pin_hash_enc,
                              new_pin_enc=new_pin_enc,
                              pin_auth=pin_auth)
+
+
+class CredentialManagement(object):
+    """Implementation of a draft specification of the Credential Management API.
+    WARNING: This specification is not final and this class is likely to change.
+
+    :param ctap: An instance of a CTAP2 object.
+    :param pin_protocol: The PIN protocol version used.
+    :param pin_token: A valid pin_token for the current CTAP session.
+    """
+
+    @unique
+    class CMD(IntEnum):
+        GET_CREDS_METADATA = 0x01
+        ENUMERATE_RPS_BEGIN = 0x02
+        ENUMERATE_RPS_NEXT = 0x03
+        ENUMERATE_CREDS_BEGIN = 0x04
+        ENUMERATE_CREDS_NEXT = 0x05
+        DELETE_CREDENTIAL = 0x06
+
+    @unique
+    class SUB_PARAMETER(IntEnum):
+        RP_ID_HASH = 0x01
+        CREDENTIAL_ID = 0x02
+
+    @unique
+    class RESULT(IntEnum):
+        EXISTING_CRED_COUNT = 0x01
+        MAX_REMAINING_COUNT = 0x02
+        RP = 0x03
+        RP_ID_HASH = 0x04
+        TOTAL_RPS = 0x05
+        USER = 0x06
+        CREDENTIAL_ID = 0x07
+        PUBLIC_KEY = 0x08
+        TOTAL_CREDENTIALS = 0x09
+        CRED_PROTECT = 0x0a
+
+    def __init__(self, ctap, pin_protocol, pin_token):
+        self.ctap = ctap
+        self.pin_protocol = pin_protocol
+        self.pin_token = pin_token
+
+    def _call(self, sub_cmd, params=None, auth=True):
+        kwargs = {
+            'sub_cmd': sub_cmd,
+            'sub_cmd_params': params
+        }
+        if auth:
+            msg = struct.pack('>B', sub_cmd)
+            if params is not None:
+                msg += cbor.encode(params)
+            kwargs['pin_protocol'] = self.pin_protocol
+            kwargs['pin_auth'] = hmac_sha256(self.pin_token, msg)[:16]
+        return self.ctap.credential_mgmt(**kwargs)
+
+    def get_metadata(self):
+        """Get credentials metadata.
+
+        This returns the existing resident credentials count, and the max
+        possible number of remaining resident credentials (the actual number of
+        remaining credentials may depend on algorithm choice, etc).
+
+        :return: A dict containing EXISTING_CRED_COUNT, and MAX_REMAINING_COUNT.
+        """
+        return self._call(CredentialManagement.CMD.GET_CREDS_METADATA)
+
+    def enumerate_rps_begin(self):
+        """Start enumeration of RP entities of resident credentials.
+
+        This will begin enumeration of stored RP entities, returning the first
+        entity, as well as a count of the total number of entities stored.
+
+        :return: A dict containing RP, RP_ID_HASH, and TOTAL_RPS.
+        """
+        return self._call(CredentialManagement.CMD.ENUMERATE_RPS_BEGIN)
+
+    def enumerate_rps_next(self):
+        """Get the next RP entity stored.
+
+        This continues enumeration of stored RP entities, returning the next
+        entity.
+
+        :return: A dict containing RP, and RP_ID_HASH.
+        """
+        return self._call(
+            CredentialManagement.CMD.ENUMERATE_RPS_NEXT,
+            auth=False
+        )
+
+    def enumerate_rps(self):
+        """Convenience method to enumerate all RPs.
+
+        See enumerate_rps_begin and enumerate_rps_next for details.
+        """
+        first = self.enumerate_rps_begin()
+        n_rps = first[CredentialManagement.RESULT.TOTAL_RPS]
+        if n_rps == 0:
+            return []
+        rest = [self.enumerate_rps_next()
+                for _ in range(
+                    1,
+                    n_rps
+                )]
+        return [first] + rest
+
+    def enumerate_creds_begin(self, rp_id_hash):
+        """Start enumeration of resident credentials.
+
+        This will begin enumeration of resident credentials for a given RP,
+        returning the first credential, as well as a count of the total number
+        of resident credentials stored for the given RP.
+
+        :param rp_id_hash: SHA256 hash of the RP ID.
+        :return: A dict containing USER, CREDENTIAL_ID, PUBLIC_KEY, and
+            TOTAL_CREDENTIALS.
+        """
+        return self._call(
+            CredentialManagement.CMD.ENUMERATE_CREDS_BEGIN,
+            {CredentialManagement.SUB_PARAMETER.RP_ID_HASH: rp_id_hash}
+        )
+
+    def enumerate_creds_next(self):
+        """Get the next resident credential stored.
+
+        This continues enumeration of resident credentials, returning the next
+        credential.
+
+        :return: A dict containing USER, CREDENTIAL_ID, and PUBLIC_KEY.
+        """
+        return self._call(
+            CredentialManagement.CMD.ENUMERATE_CREDS_NEXT,
+            auth=False
+        )
+
+    def enumerate_creds(self, *args, **kwargs):
+        """Convenience method to enumerate all resident credentials for an RP.
+
+        See enumerate_creds_begin and enumerate_creds_next for details.
+        """
+        try:
+            first = self.enumerate_creds_begin(*args, **kwargs)
+        except CtapError as e:
+            if e.code == CtapError.ERR.NO_CREDENTIALS:
+                return []
+            raise  # Other error
+        rest = [self.enumerate_creds_next()
+                for _ in range(
+                    1,
+                    first.get(CredentialManagement.RESULT.TOTAL_CREDENTIALS, 1)
+                )]
+        return [first] + rest
+
+    def delete_cred(self, cred_id):
+        """Delete a resident credential.
+
+        :param cred_id: The ID of the credential to delete.
+        """
+        return self._call(
+            CredentialManagement.CMD.DELETE_CREDENTIAL,
+            {CredentialManagement.SUB_PARAMETER.CREDENTIAL_ID: cred_id}
+        )
