@@ -197,11 +197,13 @@ class Fido2Server(object):
         if not constant_time.bytes_eq(self.rp.id_hash,
                                       attestation_object.auth_data.rp_id_hash):
             raise ValueError('Wrong RP ID hash in response.')
+        if not attestation_object.auth_data.is_user_present():
+            raise ValueError('User Present flag not set.')
 
         if state['user_verification'] is USER_VERIFICATION.REQUIRED and \
                 not attestation_object.auth_data.is_user_verified():
             raise ValueError(
-                'User verification required, but User verified flag not set.')
+                'User verification required, but User Verified flag not set.')
 
         if self.attestation != ATTESTATION.NONE:
             att_verifier = UnsupportedAttestation()
@@ -273,6 +275,8 @@ class Fido2Server(object):
             raise ValueError('Wrong challenge in response.')
         if not constant_time.bytes_eq(self.rp.id_hash, auth_data.rp_id_hash):
             raise ValueError('Wrong RP ID hash in response.')
+        if not auth_data.is_user_present():
+            raise ValueError('User Present flag not set.')
 
         if state['user_verification'] is USER_VERIFICATION.REQUIRED and \
                 not auth_data.is_user_verified():
@@ -305,16 +309,31 @@ class U2FFido2Server(Fido2Server):
     See https://www.w3.org/TR/webauthn/#sctn-appid-extension for details.
 
     :param app_id: The appId which was used for U2F registration.
+    :param verify_u2f_origin: (optional) Alternative function to validate an
+        origin for U2F credentials..
     For other parameters, see Fido2Server.
     """
 
-    def __init__(self, app_id, rp, *args, **kwargs):
+    def __init__(self, app_id, rp, verify_u2f_origin=None, *args,
+                 **kwargs):
         super(U2FFido2Server, self).__init__(rp, *args, **kwargs)
-        kwargs['verify_origin'] = lambda o: verify_app_id(app_id, o)
         kwargs['attestation_types'] = [FidoU2FAttestation()]
+        if verify_u2f_origin:
+            kwargs['verify_origin'] = verify_u2f_origin
+        else:
+            kwargs['verify_origin'] = lambda o: verify_app_id(app_id, o)
         self._app_id = app_id
         self._app_id_server = Fido2Server(RelyingParty(app_id), *args, **kwargs)
 
+    def register_begin(self, *args, **kwargs):
+        req, state = super(U2FFido2Server, self).register_begin(
+            *args,
+            **kwargs
+        )
+        req['publicKey'].setdefault('extensions', {})['appidExclude'] = \
+            self._app_id
+        return req, state
+
     def authenticate_begin(self, *args, **kwargs):
         req, state = super(U2FFido2Server, self).authenticate_begin(
             *args,
@@ -325,6 +344,7 @@ class U2FFido2Server(Fido2Server):
 
     def authenticate_complete(self, *args, **kwargs):
         try:
-            super(U2FFido2Server, self).authenticate_complete(*args, **kwargs)
+            return super(U2FFido2Server, self)\
+                .authenticate_complete(*args, **kwargs)
         except ValueError:
-            self._app_id_server.authenticate_complete(*args, **kwargs)
+            return self._app_id_server.authenticate_complete(*args, **kwargs)
